# SPDX-License-Identifier: Apache-2.0
# Standard
import operator

# Third Party
from datasets import Dataset

# Local
from .block import Block
from ..registry import BlockRegistry
from ..logger_config import setup_logger

logger = setup_logger(__name__)


@BlockRegistry.register("FilterByValueBlock")
class FilterByValueBlock(Block):
    def __init__(
        self, filter_column, filter_value, operation, convert_dtype=None, **batch_kwargs
    ) -> None:
        """
        Initializes a new instance of the FilterByValueBlock class.

        Parameters:
        - filter_column (str): The name of the column in the dataset to apply the filter on.
        - filter_value (any or list of any): The value(s) to filter by.
        - operation (callable): A function that takes two arguments (column value and filter value) and returns a boolean indicating whether the row should be included in the filtered dataset.
        - convert_dtype (callable, optional): A function to convert the data type of the filter column before applying the filter. Defaults to None.
        - **batch_kwargs: Additional kwargs for batch processing.

        Returns:
        None
        """
        super().__init__(block_name=self.__class__.__name__)
        self.value = filter_value if isinstance(filter_value, list) else [filter_value]
        self.column_name = filter_column
        self.operation = operation
        self.convert_dtype = convert_dtype
        self.num_procs = batch_kwargs.get("num_procs", 1)

    def _convert_dtype(self, sample):
        try:
            sample[self.column_name] = self.convert_dtype(sample[self.column_name])
        except ValueError as e:
            logger.error(
                "Error converting dtype: %s, filling with None to be filtered later", e
            )
            sample[self.column_name] = None
        return sample

    def generate(self, samples) -> Dataset:
        if self.convert_dtype:
            samples = samples.map(
                self._convert_dtype,
                num_proc=self.num_procs,
            )

        if self.operation == operator.contains:
            samples = samples.filter(
                lambda x: self.operation(self.value, x[self.column_name]),
                num_proc=self.num_procs,
            )

        samples = samples.filter(
            lambda x: x[self.column_name] is not None,
            num_proc=self.num_procs,
        )

        samples = samples.filter(
            lambda x: any(
                self.operation(x[self.column_name], value) for value in self.value
            ),
            num_proc=self.num_procs,
        )

        return samples



# SPDX-License-Identifier: Apache-2.0

# Third Party
import pytest

# First Party
from sdg_hub.utils import chunking

# Local
from .testdata import testdata


class TestChunking:
    """Test collection in instructlab.utils.chunking."""

    def test_chunk_docs_wc_exeeds_ctx_window(self):
        with pytest.raises(ValueError) as exc:
            chunking.chunk_document(
                documents=testdata.documents,
                chunk_word_count=1000,
                server_ctx_size=1034,
            )
        assert (
            "Given word count (1000) per doc will exceed the server context window size (1034)"
            in str(exc.value)
        )

    def test_chunk_docs_chunk_overlap_error(self):
        with pytest.raises(ValueError) as exc:
            chunking.chunk_document(
                documents=testdata.documents,
                chunk_word_count=5,
                server_ctx_size=1034,
            )
        assert (
            "Got a larger chunk overlap (100) than chunk size (24), should be smaller"
            in str(exc.value)
        )

    def test_chunk_docs_long_lines(self):
        chunk_words = 50
        chunks = chunking.chunk_document(
            documents=testdata.long_line_documents,
            chunk_word_count=chunk_words,
            server_ctx_size=4096,
        )
        max_tokens = chunking._num_tokens_from_words(chunk_words)
        max_chars = chunking._num_chars_from_tokens(max_tokens)
        max_chars += chunking._DEFAULT_CHUNK_OVERLAP  # add in the chunk overlap
        max_chars += 50  # and a bit extra for some really long words
        for chunk in chunks:
            assert len(chunk) <= max_chars



# Standard
from unittest.mock import MagicMock, patch
import os
import unittest

# Third Party
import yaml

# First Party
from sdg_hub.flow import Flow


class TestFlow(unittest.TestCase):
    def setUp(self):
        self.flow = Flow(MagicMock())

    def test_config_relative_to_flow(self):
        flow = self.flow.get_flow_from_file("tests/testdata/test_flow_1.yaml")
        block = flow[0]["block_type"](**flow[0]["block_config"])

        self.assertEqual(block.block_config["introduction"], "intro")

    def test_config_relative_to_package(self):
        with open(
            "tests/testdata/test_flow_1.yaml", "r", encoding="utf-8"
        ) as yaml_file:
            y = yaml.safe_load(yaml_file)
        y[0]["block_config"]["config_path"] = (
            "configs/skills/simple_generate_qa_freeform.yaml"
        )
        with patch("yaml.safe_load", new_callable=MagicMock) as mock_safe_load:
            mock_safe_load.return_value = y
            flow = self.flow.get_flow_from_file("tests/testdata/test_flow_1.yaml")
        block = flow[0]["block_type"](**flow[0]["block_config"])

        self.assertEqual(
            block.block_config["introduction"],
            "Develop a series of question and answer pairs to perform a task.",
        )

    def test_config_absolute(self):
        with open(
            "tests/testdata/test_flow_1.yaml", "r", encoding="utf-8"
        ) as yaml_file:
            y = yaml.safe_load(yaml_file)
        y[0]["block_config"]["config_path"] = os.path.abspath(
            "src/sdg_hub/configs/skills/simple_generate_qa_freeform.yaml"
        )
        with patch("yaml.safe_load", new_callable=MagicMock) as mock_safe_load:
            mock_safe_load.return_value = y
            flow = self.flow.get_flow_from_file("tests/testdata/test_flow_1.yaml")
        block = flow[0]["block_type"](**flow[0]["block_config"])

        self.assertEqual(
            block.block_config["introduction"],
            "Develop a series of question and answer pairs to perform a task.",
        )

    def test_config_list_mix(self):
        with open(
            "tests/testdata/test_flow_2.yaml", "r", encoding="utf-8"
        ) as yaml_file:
            y = yaml.safe_load(yaml_file)
        y[0]["block_config"]["config_paths"]["k3"] = os.path.abspath(
            "src/sdg_hub/configs/skills/simple_generate_qa_freeform.yaml"
        )

        with patch("yaml.safe_load", new_callable=MagicMock) as mock_safe_load:
            mock_safe_load.return_value = y
            flow = self.flow.get_flow_from_file("tests/testdata/test_flow_2.yaml")
        block = flow[0]["block_type"](**flow[0]["block_config"])

        self.assertEqual(block.block_config["introduction"], "intro")
        self.assertEqual(len(block.prompt_template), 3)



# SPDX-License-Identifier: Apache-2.0
# Third Party
from datasets import Dataset

# Local
from .logger_config import setup_logger

logger = setup_logger(__name__)


class EmptyDatasetError(Exception):
    pass


class Pipeline:
    def __init__(self, chained_blocks: list) -> None:
        """
        Initialize the Pipeline class with a configuration dictionary.
        config_dict: the run config py or yaml loaded into a dictionary
        """
        # pipeline config is the run configuration that consists of the pipeline steps
        self.chained_blocks = chained_blocks

    def _drop_duplicates(self, dataset, cols):
        """
        Drop duplicates from the dataset based on the columns provided.
        """
        df = dataset.to_pandas()
        df = df.drop_duplicates(subset=cols).reset_index(drop=True)
        return Dataset.from_pandas(df)

    def generate(self, dataset) -> Dataset:
        """
        Generate the dataset by running the pipeline steps.
        dataset: the input dataset
        """
        for block_prop in self.chained_blocks:
            block_type = block_prop["block_type"]
            block_config = block_prop["block_config"]
            drop_columns = block_prop.get("drop_columns", [])
            gen_kwargs = block_prop.get("gen_kwargs", {})
            drop_duplicates_cols = block_prop.get("drop_duplicates", False)
            block = block_type(**block_config)

            logger.debug("------------------------------------\n")
            logger.debug("Running block: %s", block_config["block_name"])
            logger.debug("Input dataset: %s", dataset)

            dataset = block.generate(dataset, **gen_kwargs)

            if len(dataset) == 0:
                raise EmptyDatasetError(
                    f"Pipeline stopped: Empty dataset after running block: {block_config['block_name']}"
                )

            drop_columns_in_ds = [e for e in drop_columns if e in dataset.column_names]
            if drop_columns:
                dataset = dataset.remove_columns(drop_columns_in_ds)

            if drop_duplicates_cols:
                dataset = self._drop_duplicates(dataset, cols=drop_duplicates_cols)

            logger.debug("Output dataset: %s", dataset)
            logger.debug("------------------------------------\n\n")

        return dataset



# SPDX-License-Identifier: Apache-2.0

# Standard
from enum import Enum
from typing import Any
import json
import os
import random
import re
import uuid

# Third Party
from datasets import Dataset
import yaml

# First Party
# pylint: disable=ungrouped-imports
from sdg_hub import utils
from sdg_hub.logger_config import setup_logger
from .datautils import safe_concatenate_datasets

logger = setup_logger(__name__)


class TaxonomyType(Enum):
    KNOWLEDGE = "knowledge"
    SKILL = "skill"


def _unescape(s):
    return bytes(s, "utf-8").decode("utf-8").strip()


# This is a hack because the simple workflow returns a q/a pair as a single output.
# We could possibly try to ask for them separately, but it would cost twice the inference
# API calls. All of this is because the smallest models we use on small environments
# for testing and demos weren't good enough to follow the strict formatting instructions used
# in the full pipeline.
def _get_question(synth_example: dict):
    if "question" in synth_example:
        return synth_example["question"]

    if not synth_example.get("output"):
        raise utils.GenerateException(
            f"Error: output not found in synth_example: {synth_example}"
        )

    parts = synth_example["output"].split("?", 1)
    if len(parts) != 2:
        logger.warning(f"Failed to split generated q&a: {synth_example['output']}")
    return parts[0].strip() + "?" if len(parts) == 2 else ""


# This is also a hack. See the comment above _get_question.
def _get_response(synth_example: dict):
    if "response" in synth_example:
        return synth_example["response"]

    if "output" not in synth_example:
        raise utils.GenerateException(
            f"Error: output not found in synth_example: {synth_example}"
        )

    parts = synth_example["output"].split("?", 1)
    if len(parts) != 2:
        logger.warning(f"Failed to split generated q&a: {synth_example['output']}")
    return parts[1].strip() if len(parts) == 2 else parts[0].strip()


def _convert_to_hack_fmt(sample: dict, sys_prompt: str):
    """
    Convert a sample dictionary to contain 'system', 'user', and 'assistant' columns.

    Note: We should remove this function in the future when we resolve this issue and
    standardize the format to messages.
    """
    # Create user query message
    user_query = _unescape(_get_question(sample))
    response = _unescape(_get_response(sample))
    if "context" in sample:
        user_query = f"{sample['context']}\n\n{user_query}"

    sample["id"] = str(uuid.uuid4())
    sample["system"] = sys_prompt
    sample["user"] = user_query
    sample["assistant"] = response

    return sample


def _convert_to_messages(sample: dict, sys_prompt: str):
    """
    Convert a sample dictionary to contain 'messages'
    and 'metadata' columns required for training.
    """
    # Create user query message
    user_query = _unescape(_get_question(sample))
    response = _unescape(_get_response(sample))

    sample["id"] = str(uuid.uuid4())
    sample["messages"] = [
        {"content": sys_prompt, "role": "system"},
        {"content": user_query, "role": "user"},
        {"content": response, "role": "assistant"},
    ]

    return sample


def create_auxiliary_dataset(generated_dataset: Dataset):
    if "dataset_type" not in generated_dataset.column_names:
        return None

    # get module path of the current file
    module_dir = os.path.dirname(os.path.abspath(__file__))
    aux_inst_path = os.path.join(module_dir, "../configs/knowledge/auxilary_instructions.yaml")
    if os.path.isfile(
            aux_inst_path
    ):
        with open(aux_inst_path, "r", encoding="utf-8") as fp:
            auxiliary_inst = yaml.safe_load(fp)
    else:
        logger.error(f"auxiliary instructions file not found at {aux_inst_path}")
        return None
    auxiliary_ds = generated_dataset.filter(
        lambda x: x["dataset_type"] != "base_document"
    )
    unique_document_auxiliary = auxiliary_ds.to_pandas().drop_duplicates(
        subset=["document"]
    )
    unique_document_auxiliary = Dataset.from_pandas(unique_document_auxiliary)
    unique_document_auxiliary = unique_document_auxiliary.remove_columns(
        [
            col
            for col in unique_document_auxiliary.column_names
            if col
            not in [
                "raw_document",
                "document_outline",
                "domain",
                "dataset_type",
                "document",
            ]
        ]
    )
    unique_document_auxiliary = unique_document_auxiliary.rename_columns(
        {"raw_document": "context", "document": "response"}
    )

    def __create_auxiliary_ds(rec):
        instruction = random.choice(auxiliary_inst[rec["dataset_type"]])
        messages = [
            {"role": "user", "content": f"{rec['context']}\n\n{instruction}"},
            {"role": "assistant", "content": rec["response"]},
        ]
        metadata = json.dumps(
            {
                "dataset_type": rec["dataset_type"],
                "raw_document": rec["context"],
                "dataset": f"document_{rec['dataset_type']}",
                "domain": rec["domain"],
            }
        )
        return {"messages": messages, "metadata": metadata, "id": str(uuid.uuid4())}

    unique_document_auxiliary = unique_document_auxiliary.map(
        __create_auxiliary_ds, remove_columns=unique_document_auxiliary.column_names
    )
    return unique_document_auxiliary


def generate_knowledge_qa_dataset(
    generated_dataset: Dataset, keep_context_separate=False
):
    def __create_qa_row(rec):
        context = rec["document"]
        instruction = rec["question"]
        response = rec["response"]
        metadata = {
            "sdg_document": rec["document"],
            "domain": rec["domain"],
            "dataset": "document_knowledge_qa",
        }
        if "raw_document" in rec and "dataset_type" in rec:
            metadata.update(
                {
                    "raw_document": rec["raw_document"],
                    "dataset_type": rec["dataset_type"],
                }
            )
        metadata = json.dumps(metadata)
        if keep_context_separate:
            messages = [
                {"role": "user", "content": f"{instruction}"},
                {"role": "assistant", "content": response},
            ]
            return {
                "messages": messages,
                "metadata": metadata,
                "id": str(uuid.uuid4()),
                "context": context,
            }
        else:
            messages = [
                {"role": "user", "content": f"{context}\n\n{instruction}"},
                {"role": "assistant", "content": response},
            ]

            return {"messages": messages, "metadata": metadata, "id": str(uuid.uuid4())}

    knowledge_ds = generated_dataset.map(
        __create_qa_row, remove_columns=generated_dataset.column_names
    )
    return knowledge_ds


def build_raft_dataset(ds: Dataset, p, num_doc_in_context=4):
    all_context = list(set(ds["context"]))

    def _pick_documents(rec, p):
        answer_document = rec["context"]
        selected_docs = [e for e in all_context if e != answer_document]
        if len(selected_docs) > 0:
            if len(selected_docs) < num_doc_in_context:
                logger.info(
                    f"Number of unique document is {len(selected_docs)} which is less than {num_doc_in_context}. Using all the documents in the RAFT context"
                )
            if random.uniform(0, 1) < p:
                # golden/answer + distractor documents                                                                                         
                docs = (
                    random.sample(selected_docs, k=num_doc_in_context-1) + [answer_document]
                    if len(selected_docs) >= (num_doc_in_context-1)
                    else selected_docs + [answer_document]
                )
            else:
                # distractor documents                                                                                                         
                docs = (
                    random.sample(selected_docs, k=num_doc_in_context)
                    if len(selected_docs) >= num_doc_in_context
                    else selected_docs
                )
        else:
            logger.info("Only 1 unique document found. Turning off RAFT styling")
            docs = [answer_document]

        random.shuffle(docs)

        docs = "\n".join(([f"Document:\n{e}\n\n" for idx, e in enumerate(docs)]))
        user_idx, user_msg = [
            (idx, rec_msg)
            for idx, rec_msg in enumerate(rec["messages"])
            if rec_msg["role"] == "user"
        ][0]
        user_inst = user_msg["content"]
        rec["messages"][user_idx]["content"] = f"{docs}\n\n{user_inst}"
        rec["messages"] = rec["messages"]
        metadata = json.loads(rec["metadata"])
        metadata["dataset"] += f"_raft_p{p}"
        rec["metadata"] = json.dumps(metadata)
        return rec

    ds = ds.map(_pick_documents, fn_kwargs={"p": p} , remove_columns=["context"])
    return ds



def _conv_pretrain(rec):
    rec["messages"] = [
        {
            "role": "pretraining",
            "content": f"<|user|>\n{rec['messages'][0]['content']}\n<|assistant|>\n{rec['messages'][1]['content']}",
        }
    ]
    return rec


def create_knowledge_regular_ds(generated_dataset: Dataset):
    # Phase 1.0
    knowledge_ds = generate_knowledge_qa_dataset(
        generated_dataset, keep_context_separate=True
    )
    knowledge_ds = build_raft_dataset(knowledge_ds, p=0.4)

    auxiliary_dataset = create_auxiliary_dataset(generated_dataset)
    if auxiliary_dataset is not None:
        transformed_data = safe_concatenate_datasets([knowledge_ds, auxiliary_dataset])
    else:
        transformed_data = knowledge_ds
    return transformed_data


def create_knowledge_pretraining_ds(generated_dataset: Dataset):
    # Phase 0.7
    knowledge_ds = generate_knowledge_qa_dataset(
        generated_dataset, keep_context_separate=False
    )
    knowledge_ds = knowledge_ds.map(_conv_pretrain)

    auxiliary_dataset = create_auxiliary_dataset(generated_dataset)
    if auxiliary_dataset is not None:
        auxiliary_dataset = auxiliary_dataset.map(_conv_pretrain)
        transformed_data = safe_concatenate_datasets([knowledge_ds, auxiliary_dataset])
    else:
        transformed_data = knowledge_ds
    return transformed_data


def post_process_mcq(ds: Dataset, is_mmlu_eval: bool = False) -> Dataset:
    """Filters out badly generated data, adds dataset type column

    Args:
        ds (Dataset): mcq generated dataset from mmmlu pipeline
        is_mmlu_eval (bool, optional): _description_. Defaults to False.

    Returns:
        Dataset: Hf Dataset with new column, filtered dataset
    """
    ds = ds.filter(lambda x: ")" in x["mmlubench_answer"])
    ds = ds.filter(lambda x: "A)" in x["mmlubench_question"])
    ds = ds.add_column("dataset_type", ["mcq_qa"] * ds.num_rows)
    if is_mmlu_eval:
        return format_mmlu_style(ds)
    return ds


def extract_options(text: str) -> list[Any]:
    """regex to extract options from mcq

    Args:
        text (str): question with options/mcq choices

    Returns:
        list[Any]: options under question that match the pattern.
    """
    # Use a regular expression to find patterns and capture the text after the letter and parenthesis
    pattern = r"\b[A-Z]\) (.+)"
    matches = re.findall(pattern, text)
    return matches


def format_mmlu_style(ds: Dataset) -> Dataset:
    """Format the dataset according to lm-harness mmlu requirement.

    Args:
        ds (Dataset): input dataset

    Returns:
        Dataset: formated hf dataset
    """
    ds = ds.map(
        lambda x: {"answer": x["mmlubench_answer"][: x["mmlubench_answer"].index(")")]}
    )
    ds = ds.map(lambda x: {"choices": extract_options(x["mmlubench_question"])})
    ds = ds.map(
        lambda x: {
            "question": x["mmlubench_question"][
                : x["mmlubench_question"].index("A)")
            ].strip()
        }
    )
    ds = ds.rename_columns({"domain": "subject"})
    ds = ds.filter(lambda x: x["choices"])
    ds = ds.filter(lambda x: len(x["choices"]) == 4)
    ds = ds.filter(lambda x: x["answer"] in ["A", "B", "C", "D"])
    ds = ds.class_encode_column("answer")
    return ds


def create_mmlu_evaluation_dataset(generate_mcq_dataset: Dataset) -> Dataset:
    """Filter, format and return mcq dataset that is compatible with lm-harness for doing mmlu-style evaluation

    Args:
        generate_mcq_dataset (Dataset): sdg generated mcq dataset
    Returns:
        Dataset: MMLU MCQ datast
    """
    mmlu_dataset = post_process_mcq(generate_mcq_dataset, is_mmlu_eval=True)
    return mmlu_dataset


def create_mmlu_evaluation_yaml(task_name, eval_data_file_path, yaml_file_path):
    """
    Prepare Task Yaml that will be used in lm_eval_harness to evaluate knowledge using mmlu style metric
    """
    task_yaml = {
        "task": task_name,
        "dataset_kwargs": {"data_files": {"test": eval_data_file_path}},
        "include": "_default_mmlu_pr_template_yaml",
        "group": "mmlu_pr",
    }
    with open(yaml_file_path, "w", encoding="utf-8") as yaml_file:
        yaml.dump(task_yaml, yaml_file, default_flow_style=False)



# SPDX-License-Identifier: Apache-2.0
# Standard
from abc import ABC
from collections import ChainMap
from typing import Any, Dict, Union

# Third Party
from jinja2 import Template, UndefinedError
import yaml

# Local
from ..registry import BlockRegistry
from ..logger_config import setup_logger

logger = setup_logger(__name__)


@BlockRegistry.register("Block")
class Block(ABC):
    def __init__(self, block_name: str) -> None:
        self.block_name = block_name

    @staticmethod
    def _validate(prompt_template: Template, input_dict: Dict[str, Any]) -> bool:
        """
        Validate the input data for this block. This method validates whether all required
        variables in the Jinja template are provided in the input_dict.

        :param prompt_template: The Jinja2 template object.
        :param input_dict: A dictionary of input values to check against the template.
        :return: True if the input data is valid (i.e., no missing variables), False otherwise.
        """
        
        class Default(dict):
            def __missing__(self, key: str) -> None:
                raise KeyError(key)
        
        try:
            # Try rendering the template with the input_dict
            prompt_template.render(ChainMap(input_dict, Default()))
            return True
        except UndefinedError as e:
            logger.error(f"Missing key: {e}")
            return False

    def _load_config(self, config_path: str) -> Union[Dict[str, Any], None]:
        """
        Load the configuration file for this block.

        :param config_path: The path to the configuration file.
        :return: The loaded configuration.
        """
        with open(config_path, "r", encoding="utf-8") as config_file:
            return yaml.safe_load(config_file)



# Third Party
from datasets import Dataset

# Local
from .block import Block
from ..registry import BlockRegistry
from ..logger_config import setup_logger

logger = setup_logger(__name__)


@BlockRegistry.register("IterBlock")
class IterBlock(Block):
    def __init__(self, block_name, num_iters, block_type, block_kwargs, **kwargs):
        super().__init__(block_name)
        self.num_iters = num_iters
        self.block = block_type(**block_kwargs)
        self.gen_kwargs = kwargs.get("gen_kwargs", {})
        self.gen_kwargs = kwargs.get("gen_kwargs", {})

    def generate(self, samples, **gen_kwargs) -> Dataset:
        generated_samples = []
        num_iters = self.num_iters

        for _ in range(num_iters):
            batch_generated = self.block.generate(
                samples, **{**self.gen_kwargs, **gen_kwargs}
            )
            generated_samples.extend(batch_generated)

        return Dataset.from_list(generated_samples)



# SPDX-License-Identifier: Apache-2.0
# Third Party
from datasets import Dataset

# Local
from .block import Block
from ..registry import BlockRegistry
from ..logger_config import setup_logger

logger = setup_logger(__name__)


@BlockRegistry.register("SamplePopulatorBlock")
class SamplePopulatorBlock(Block):
    def __init__(self, config_paths, column_name, post_fix="", **batch_kwargs) -> None:
        super().__init__(
            block_name=self.__class__.__name__
        )  # Call the base class's __init__
        self.configs = {}
        for config in config_paths:
            if post_fix:
                config_name = config.replace(".yaml", f"_{post_fix}.yaml")
            else:
                config_name = config
            config_key = config.split("/")[-1].split(".")[0]
            self.configs[config_key] = self._load_config(config_name)
        self.column_name = column_name
        self.num_procs = batch_kwargs.get("num_procs", 8)

    def _generate(self, sample) -> dict:
        sample = {**sample, **self.configs[sample[self.column_name]]}
        return sample

    def generate(self, samples) -> Dataset:
        samples = samples.map(self._generate, num_proc=self.num_procs)
        return samples


@BlockRegistry.register("SelectorBlock")
class SelectorBlock(Block):
    def __init__(self, choice_map, choice_col, output_col, **batch_kwargs) -> None:
        super().__init__(block_name=self.__class__.__name__)
        self.choice_map = choice_map
        self.choice_col = choice_col
        self.output_col = output_col
        self.num_procs = batch_kwargs.get("num_procs", 8)

    def _generate(self, sample) -> dict:
        sample[self.output_col] = sample[self.choice_map[sample[self.choice_col]]]
        return sample

    def generate(self, samples: Dataset) -> Dataset:
        samples = samples.map(self._generate, num_proc=self.num_procs)
        return samples


@BlockRegistry.register("CombineColumnsBlock")
class CombineColumnsBlock(Block):
    def __init__(self, columns, output_col, separator="\n\n", **batch_kwargs) -> None:
        super().__init__(block_name=self.__class__.__name__)
        self.columns = columns
        self.output_col = output_col
        self.separator = separator
        self.num_procs = batch_kwargs.get("num_procs", 8)

    def _generate(self, sample) -> dict:
        sample[self.output_col] = self.separator.join(
            [sample[col] for col in self.columns]
        )
        return sample

    def generate(self, samples: Dataset) -> Dataset:
        samples = samples.map(self._generate, num_proc=self.num_procs)
        return samples


@BlockRegistry.register("FlattenColumnsBlock")
class FlattenColumnsBlock(Block):
    def __init__(self, block_name: str, var_cols: list, value_name: str, var_name: str) -> None:
        super().__init__(block_name=block_name)
        self.var_cols = var_cols
        self.value_name = value_name
        self.var_name = var_name

    def generate(self, samples: Dataset) -> Dataset:
        df = samples.to_pandas()
        id_cols = [col for col in samples.column_names if col not in self.var_cols]
        flatten_df = df.melt(id_vars=id_cols, 
                             value_vars=self.var_cols, 
                             value_name=self.value_name,
                             var_name=self.var_name)
        
        return Dataset.from_pandas(flatten_df)


@BlockRegistry.register("DuplicateColumns")
class DuplicateColumns(Block):
    def __init__(self, block_name: str, columns_map: dict) -> None:
        """Create duplicate of columns specified in column map.

        Args:
            columns_map (dict): mapping of existing column to new column names
        """
        super().__init__(block_name=block_name)
        self.columns_map = columns_map    
    
    
    def generate(self, samples: Dataset):
        for col_to_dup in self.columns_map:
            samples = samples.add_column(self.columns_map[col_to_dup], samples[col_to_dup])
        return samples


@BlockRegistry.register("RenameColumns")
class RenameColumns(Block):
    def __init__(self, block_name: str, columns_map: dict) -> None:
        """Rename dataset columns.

        Args:
            columns_map (dict): mapping of existing column to new column names
        """
        self.columns_map = columns_map
        super().__init__(block_name=block_name)
    
    
    def generate(self, samples: Dataset):
        samples = samples.rename_columns(self.columns_map)
        return samples


@BlockRegistry.register("SetToMajorityValue")
class SetToMajorityValue(Block):
    def __init__(self, block_name: str, col_name) -> None:
        self.col_name = col_name
        super().__init__(block_name)
    
    def generate(self, samples: Dataset):
        samples = samples.to_pandas()
        samples[self.col_name] = samples[self.col_name].mode()[0]
        return Dataset.from_pandas(samples)


# Standard
from pathlib import Path
import json

# Third Party
from datasets import Dataset
from tabulate import tabulate
from transformers import AutoTokenizer
import yaml

# First Party
from sdg_hub.logger_config import setup_logger

# Local
from .datautils import safe_concatenate_datasets
from .chunking import chunk_document

logger = setup_logger(__name__)


def fuse_texts(text_list, short_length_threshold=100):
    fused_texts = []
    previous_long_text = ""

    for text in text_list:
        word_count = len(text.split())

        if word_count <= short_length_threshold and previous_long_text:
            # Append the short text to the last long text
            fused_texts[-1] += "\n\n" + text
        else:
            # This is a long text, so add it to the list and remember it
            fused_texts.append(text)
            previous_long_text = text

    return fused_texts


def handle_footnote(book_element):
    pass


def create_tokenizer():
    return AutoTokenizer.from_pretrained("instructlab/granite-7b-lab")


def get_token_count(text, tokenizer):
    return len(tokenizer.tokenize(text))


def add_heading_formatting(text):
    text = text.split(".")
    # TODO: Change this from hardcoded to something that makes sense
    if len(text) > 1 and len(text[0].split(" ")) < 3:
        text = f"**{text[0]}**" + ".".join(text[1:])
    else:
        text = ".".join(text)
    return text


def generate_table_from_parsed_rep(item):
    """
    Generate the table from the parsed representation and return
    """
    caption = ""
    if "text" in item:
        # print("caption: ", item["text"])
        caption = item["text"]

    data = item["data"]

    if len(data) <= 1 or len(data[0]) <= 1:
        return ""

    table = []
    for i, row in enumerate(data):
        trow = []
        for j, cell in enumerate(row):
            trow.append(cell["text"])
        table.append(trow)

    table_text = tabulate(table, tablefmt="github")
    if caption:
        table_text += f"\nCaption: {caption}\n"
    return table_text


def get_table(json_book, table_ref):
    parts = table_ref.split("/")
    table_text = generate_table_from_parsed_rep(json_book[parts[1]][int(parts[2])])
    return table_text


def get_table_page_number(json_book, idx):
    # Get previous page number
    prev_page_num, next_page_num = None, None
    for book_element in json_book["main-text"][idx - 1 :: -1]:
        if "prov" in book_element:
            prev_page_num = book_element["prov"][0]["page"]
            break
    for book_element in json_book["main-text"][idx:]:
        if "prov" in book_element:
            next_page_num = book_element["prov"][0]["page"]
            break
    if prev_page_num is not None and next_page_num is not None:
        if prev_page_num == next_page_num:
            return prev_page_num
        else:
            return next_page_num
    elif prev_page_num is not None:
        return prev_page_num
    elif next_page_num is not None:
        return next_page_num


def build_chunks_from_docling_json(
    json_book,
    max_token_per_chunk,
    tokenizer,
    keep_same_page_thing_together=False,
    chunking_criteria=None,
):
    current_buffer = []
    document_chunks = []
    prev_page_number = None
    book_title = None

    for idx, book_element in enumerate(json_book["main-text"]):
        if book_element["type"] in [
            "page-footer",
            "picture",
            "reference",
            "meta-data",
            "figure",
            "page-header",
        ]:
            continue
        elif book_element["type"] == "footnote":
            handle_footnote(book_element)
            current_book_page_number = book_element["prov"][0]["page"]
        elif book_element["type"] in [
            "subtitle-level-1",
            "paragraph",
            "table",
            "title",
            "equation",
        ]:  # 'page-header',
            if book_element["type"] == "table":
                current_book_page_number = get_table_page_number(json_book, idx)
            else:
                current_book_page_number = book_element["prov"][0]["page"]
                book_text = book_element["text"]

            if book_element["type"] == "subtitle-level-1":
                if book_title is None:
                    book_title = book_text
                    book_text = f"# Title: **{book_text}**"
                else:
                    book_text = f"## **{book_text}**"

            if book_element["type"] == "title":
                book_text = f"# **{book_text}**"
            if book_element["type"] == "page-header":
                book_text = f"Page Header: **{book_text}**\n\n"

            if chunking_criteria is not None:
                # custom break function that can be used to chunk document
                if chunking_criteria(book_text):
                    document_chunks.append("\n\n".join(current_buffer))
                    current_buffer = []
            elif (
                prev_page_number is not None
                and prev_page_number != current_book_page_number
            ) and keep_same_page_thing_together:
                document_chunks.append("\n\n".join(current_buffer))
                current_buffer = []
            else:
                if (
                    get_token_count("\n\n".join(current_buffer), tokenizer)
                    >= max_token_per_chunk
                    and len(current_buffer) > 1
                ):
                    # chunk_text = '\n\n'.join(current_buffer[:-1])
                    # print(f"Current chunk size {get_token_count(chunk_text, tokenizer)} and max is {max_token_per_chunk}")
                    document_chunks.append("\n\n".join(current_buffer[:-1]))

                    if (
                        get_token_count(current_buffer[-1], tokenizer)
                        >= max_token_per_chunk
                    ):
                        # print(f"This is too big document to be left in the current buffer { get_token_count(current_buffer[-1], tokenizer)}")
                        document_chunks.append(current_buffer[-1])
                        current_buffer = []
                    else:
                        current_buffer = current_buffer[-1:]

            if book_element["type"] == "paragraph":
                book_text = add_heading_formatting(book_text)
            elif book_element["type"] == "table":
                book_text = get_table(json_book, book_element["$ref"])
            if "## References" in book_text or "## Acknowledgements" in book_text:
                # For reasearch papers we ignore everything after this sections
                break
            current_buffer.append(book_text)

        try:
            prev_page_number = current_book_page_number
        except:
            logger.error(book_element)
    if "\n\n".join(current_buffer) not in document_chunks:
        document_chunks.append("\n\n".join(current_buffer))
    return document_chunks


class DocProcessor:
    def __init__(
        self,
        parsed_doc_dir: Path,
        tokenizer: str = "instructlab/granite-7b-lab",
        user_config_path: Path = None,
    ):
        self.parsed_doc_dir = self._path_validator(parsed_doc_dir)
        self.user_config = self._load_user_config(
            self._path_validator(user_config_path)
        )
        self.docling_jsons = list(self.parsed_doc_dir.glob("*.json"))
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)

    def _path_validator(self, path) -> Path:
        """
        Validate the path and return a Path object.
        Args:
            path (str): Path to be validated.
        Returns:
            Path`: Path object.
        """
        if isinstance(path, str):
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f"{path} does not exist.")
        return path

    def _load_user_config(self, user_config_path: Path) -> dict:
        """
        Load the user config file.
        Args:
            user_config_path (Path): Path to the user config file.
        Returns:
            dict: User config dictionary.
        """
        # load user config as yaml
        with open(user_config_path, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)

    def _process_parsed_docling_json(self, json_fp: Path) -> Dataset:
        """
        Process the parsed docling json file and return a dataset.
        Args:
            json_fp (str): Path to the parsed docling json file.
        Returns:
            Dataset: Dataset object.
        """
        logger.info(f"Processing parsed docling json file: {json_fp}")
        with open(json_fp, "r", encoding="utf-8") as f:
            data = json.load(f)

        file_name = json_fp.name.split(".")[0]
        chunks = build_chunks_from_docling_json(
            data,
            max_token_per_chunk=500,
            tokenizer=self.tokenizer,
        )
        chunks = fuse_texts(chunks, 200)
        return Dataset.from_dict(
            {
                "document": chunks,
                "document_outline": [self.user_config["document_outline"]]
                * len(chunks),
                "document_title": [file_name] * len(chunks),
                "domain": [self.user_config["domain"]] * len(chunks),
            }
        )

    def _add_icls(self, chunked_document: Dataset) -> Dataset:
        """
        Add the ICLS label to the dataset.
        Args:
            dataset (Dataset): Dataset object.
        Returns:
            Dataset: Dataset object with ICLS label.
        """
        icl = self.user_config["seed_examples"]
        chunked_document_all_icl = []
        for icl_ in icl:
            chunked_document_all_icl.append(
                chunked_document.map(
                    lambda x: {
                        "icl_document": icl_["context"],
                        "icl_query_1": icl_["questions_and_answers"][0]["question"],
                        "icl_response_1": icl_["questions_and_answers"][0]["answer"],
                        "icl_query_2": icl_["questions_and_answers"][1]["question"],
                        "icl_response_2": icl_["questions_and_answers"][1]["answer"],
                        "icl_query_3": icl_["questions_and_answers"][2]["question"],
                        "icl_response_3": icl_["questions_and_answers"][2]["answer"],
                    }
                )
            )
        chunked_document_all_icl = safe_concatenate_datasets(chunked_document_all_icl)
        chunked_document_all_icl = chunked_document_all_icl.map(
            lambda x: {
                "chunks": chunk_document(
                    [x["document"]], server_ctx_size=4096, chunk_word_count=1024
                )
                if get_token_count(x["document"], self.tokenizer) > 1024
                else [x["document"]]
            }
        )
        df = chunked_document_all_icl.to_pandas()
        df_exploded = df.explode("chunks").reset_index(drop=True)
        new_ds = Dataset.from_pandas(df_exploded)
        new_ds = new_ds.remove_columns("document").rename_columns(
            {"chunks": "document"}
        )

        # Only keep document greater than 100 tokens
        new_ds = new_ds.filter(
            lambda x: get_token_count(x["document"], self.tokenizer) > 100
        )
        return new_ds

    def get_processed_dataset(self) -> Dataset:
        """
        Process all the parsed docling json files and return a dataset.
        Returns:
            Dataset: Dataset object.
        """
        datasets = []
        for json_fp in self.docling_jsons:
            chunk_ds = self._process_parsed_docling_json(json_fp)
            chunk_ds_with_icls = self._add_icls(chunk_ds)
            datasets.append(chunk_ds_with_icls)
        return safe_concatenate_datasets(datasets)
    
    def get_processed_markdown_dataset(self, list_md_files: list[Path]) -> Dataset:
        chunks_mds = []
        for md_file in list_md_files:
            with open(md_file, "r", encoding="utf-8") as f:
                text = f.read()
                chunks_mds.append({
            "document": text,
            "document_outline": self.user_config["document_outline"],
            "document_title": md_file,
            "domain": self.user_config["domain"],
            })
        chunk_ds = Dataset.from_list(chunks_mds)
        chunk_ds_with_icls = self._add_icls(chunk_ds)
        return chunk_ds_with_icls


# SPDX-License-Identifier: Apache-2.0

# Standard
from typing import List
import logging
import re

# Third Party
from langchain_text_splitters import Language, RecursiveCharacterTextSplitter

_DEFAULT_CHUNK_OVERLAP = 100

logger = logging.getLogger(__name__)


def _num_tokens_from_words(num_words) -> int:
    return int(num_words * 1.3)  # 1 word ~ 1.3 token


def _num_chars_from_tokens(num_tokens) -> int:
    return int(num_tokens * 4)  # 1 token ~ 4 English character


def chunk_document(documents: List, server_ctx_size, chunk_word_count) -> List[str]:
    """
    Iterates over the documents and splits them into chunks based on the word count provided by the user.
    Args:
        documents (list): List of documents retrieved from git (can also consist of a single document).
        server_ctx_size (int): Context window size of server.
        chunk_word_count (int): Maximum number of words to chunk a document.
    Returns:
         List[str]: List of chunked documents.
    """

    # Checks for input type error
    if isinstance(documents, str):
        documents = [documents]

    elif not isinstance(documents, list):
        raise TypeError(
            "Expected: documents to be a list, but got {}".format(type(documents))
        )

    no_tokens_per_doc = _num_tokens_from_words(chunk_word_count)
    if no_tokens_per_doc > int(server_ctx_size - 1024):
        raise ValueError(
            "Error: {}".format(
                str(
                    f"Given word count ({chunk_word_count}) per doc will exceed the server context window size ({server_ctx_size})"
                )
            )
        )
    # Placeholder for params
    content = []
    chunk_size = _num_chars_from_tokens(no_tokens_per_doc)
    chunk_overlap = _DEFAULT_CHUNK_OVERLAP

    # Using Markdown as default, document-specific chunking will be implemented in seperate pr.
    text_splitter = RecursiveCharacterTextSplitter.from_language(
        language=Language.MARKDOWN,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
    )

    # Determine file type for heuristics, default with markdown
    for docs in documents:
        # Use regex to remove unnecessary dashes in front of pipe characters in a markdown table.
        docs = re.sub(r"-{2,}\|", "-|", docs)
        # Remove unnecessary spaces in front of pipe characters in a markdown table.
        docs = re.sub(r"\  +\|", " |", docs)
        temp = text_splitter.create_documents([docs])
        content.extend([item.page_content for item in temp])
    return content



# SPDX-License-Identifier: Apache-2.0
# Standard
from abc import ABC
from importlib import resources
from typing import Optional
import operator
import os

# Third Party
import yaml

# Local
from .registry import BlockRegistry, PromptRegistry
from . import prompts
from . import blocks


OPERATOR_MAP = {
    "operator.eq": operator.eq,
    "operator.ge": operator.ge,
    "operator.contains": operator.contains,
}

CONVERT_DTYPE_MAP = {
    "float": float,
    "int": int,
}


class Flow(ABC):
    def __init__(
        self,
        llm_client,
        num_samples_to_generate: Optional[int] = None,
    ) -> None:
        self.llm_client = llm_client
        self.num_samples_to_generate = num_samples_to_generate
        self.base_path = str(resources.files(__package__))
        self.registered_blocks = BlockRegistry.get_registry()

    def _getFilePath(self, dirs, filename):
        """
        Find a named configuration file.

        Files are checked in the following order
            - absulute path is always used
            - checked relative to the directories in "dirs"
            - relative the the current directory

        Args:
            dirs (list): Directories in which to search for "config_path"
            config_path (str): The path to the configuration file.

        Returns:
            Selected file path
        """
        if os.path.isabs(filename):
            return filename
        for d in dirs:
            full_file_path = os.path.join(d, filename)
            if os.path.isfile(full_file_path):
                return full_file_path
        # If not found above then return the path unchanged i.e.
        # assume the path is relative to the current directory
        return filename

    def get_flow_from_file(self, yaml_path: str) -> list:
        yaml_path_relative_to_base = os.path.join(self.base_path, yaml_path)
        if os.path.isfile(yaml_path_relative_to_base):
            yaml_path = yaml_path_relative_to_base
        yaml_dir = os.path.dirname(yaml_path)

        try:
            with open(yaml_path, "r", encoding="utf-8") as yaml_file:
                flow = yaml.safe_load(yaml_file)
        except FileNotFoundError as exc:
            raise FileNotFoundError(f"File not found: {yaml_path}") from exc

        # update config with class instances
        for block in flow:
            # check if theres an llm block in the flow
            if "LLM" in block["block_type"]:
                block["block_config"]["client"] = self.llm_client
                # model_id and prompt templates
                # try to get a template using the model_id, but if model_prompt_template is provided, use that
                if block["block_config"].get("model_prompt", None) is None:
                    # try to find a match in the registry
                    matched_prompt = next(
                        (
                            key
                            for key in PromptRegistry.get_registry()
                            if key in block["block_config"]["model_id"]
                        ),
                        None,
                    )
                    if matched_prompt is not None:
                        block["block_config"]["model_prompt"] = matched_prompt
                    else:
                        raise KeyError(
                            f"Prompt not found in registry: {block['block_config']['model_id']}"
                        )

                if self.num_samples_to_generate is not None:
                    block["num_samples"] = self.num_samples_to_generate

            # update block type to llm class instance
            try:
                block["block_type"] = self.registered_blocks[block["block_type"]]
            except KeyError as exc:
                raise KeyError(
                    f"Block not found in registry: {block['block_type']}"
                ) from exc

            # update config path to absolute path
            if "config_path" in block["block_config"]:
                block["block_config"]["config_path"] = self._getFilePath(
                    [yaml_dir, self.base_path], block["block_config"]["config_path"]
                )

            # update config paths to absolute paths - this might be a list or a dict
            if "config_paths" in block["block_config"]:
                if isinstance(block["block_config"]["config_paths"], dict):
                    for key, path in block["block_config"]["config_paths"].items():
                        block["block_config"]["config_paths"][key] = self._getFilePath(
                            [yaml_dir, self.base_path], path
                        )

                elif isinstance(block["block_config"]["config_paths"], list):
                    for i, path in enumerate(block["block_config"]["config_paths"]):
                        block["block_config"]["config_paths"][i] = self._getFilePath(
                            [yaml_dir, self.base_path], path
                        )

            if "operation" in block["block_config"]:
                block["block_config"]["operation"] = OPERATOR_MAP[
                    block["block_config"]["operation"]
                ]

            if "convert_dtype" in block["block_config"]:
                block["block_config"]["convert_dtype"] = CONVERT_DTYPE_MAP[
                    block["block_config"]["convert_dtype"]
                ]

        return flow



# Third Party
from datasets import load_dataset
from openai import OpenAI
import click
import os

# First Party
from sdg_hub.flow import Flow
from sdg_hub.logger_config import setup_logger
from sdg_hub.pipeline import Pipeline
from sdg_hub.sdg import SDG


logger = setup_logger(__name__)


@click.command()
@click.option(
    "--ds_path",
    type=click.Path(exists=True),
    required=True,
    help="Path to the dataset.",
)
@click.option("--bs", type=int, default=8, show_default=True, help="Batch size.")
@click.option("--num_workers", type=int, default=32, show_default=True, help="Number of workers.")
@click.option("--save_path", type=click.Path(), required=True, help="Path to save the output.")
@click.option("--endpoint", type=str, required=True, help="Endpoint for data processing.")
@click.option(
    "--flow",
    type=click.Path(exists=True),
    required=True,
    help="Flow configuration for the process.",
)
@click.option(
    "--checkpoint_dir",
    type=click.Path(),
    required=True,
    help="Path to save checkpoints.",
)
@click.option(
    "--save_freq",
    type=int,
    default=2,
    show_default=True,
    help="Frequency to save checkpoints.",
)
@click.option("--debug", is_flag=True, help="Enable debug mode.")
def main(
    ds_path,
    bs,
    num_workers,
    save_path,
    endpoint,
    flow,
    checkpoint_dir,
    save_freq,
    debug,
):
    """
    Main function to process the dataset.

    Parameters:
    ds_path (str): Path to the dataset.
    bs (int): Batch size.
    num_workers (int): Number of workers.
    save_path (str): Path to save the output.
    endpoint (str): Endpoint for data processing.
    flow (str): Flow configuration for the process.
    checkpoint_dir (str): Path to save checkpoints.
    save_freq (int): Frequency to save checkpoints.
    debug (bool): Enable debug mode.
    """
    logger.info(f"Generation configuration: {locals()}\n\n")
    ds = load_dataset("json", data_files=ds_path, split="train")

    if debug:
        ds = ds.shuffle(seed=42).select(range(30))
        logger.info("Debug mode enabled. Using a subset of the dataset.")

    openai_api_key = os.environ.get("OPENAI_API_KEY", "EMPTY")
    openai_api_base = endpoint

    client = OpenAI(
        api_key=openai_api_key,
        base_url=openai_api_base,
    )

    if not os.path.exists(flow):
        raise FileNotFoundError(f"Flow file not found: {flow}")

    flow_cfg = Flow(client).get_flow_from_file(flow)
    sdg = SDG(
        [Pipeline(flow_cfg)],
        num_workers=num_workers,
        batch_size=bs,
        save_freq=save_freq,
    )

    generated_data = sdg.generate(ds, checkpoint_dir=checkpoint_dir)
    generated_data.to_json(save_path, orient="records", lines=True)
    logger.info(f"Data saved to {save_path}")


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



# SPDX-License-Identifier: Apache-2.0

documents = [
    """Knowledge is an awareness of facts, a familiarity with individuals and situations, 
      or a practical skill. Knowledge of facts, also called propositional knowledge, is often characterized 
      as true belief that is distinct from opinion or guesswork by virtue of justification. 
      While there is wide agreement among philosophers that propositional knowledge is a form of true belief, 
      many controversies focus on justification. This includes questions like how to understand justification, 
      whether it is needed at all, and whether something else besides it is needed. 
      These controversies intensified in the latter half of the 20th century due to a series of thought experiments 
      called Gettier cases that provoked alternative definitions."""
]

long_line_documents = [
    (
        "Knowledge is an awareness of facts, a familiarity with individuals and situations,"
        "or a practical skill. Knowledge of facts, also called propositional knowledge, is often characterized "
        "as true belief that is distinct from opinion or guesswork by virtue of justification. "
        "While there is wide agreement among philosophers that propositional knowledge is a form of true belief, "
        "many controversies focus on justification. This includes questions like how to understand justification, "
        "whether it is needed at all, and whether something else besides it is needed. "
        "These controversies intensified in the latter half of the 20th century due to a series of thought experiments "
        "called Gettier cases that provoked alternative definitions."
    )
]



# Standard
from typing import Union, List, Dict

# Third Party
from jinja2 import Template

# Local
from .logger_config import setup_logger

logger = setup_logger(__name__)


class BlockRegistry:
    """Registry for block classes to avoid manual additions to block type map."""

    _registry: Dict[str, type] = {}

    @classmethod
    def register(cls, block_name: str):
        """
        Decorator to register a block class under a specified name.

        :param block_name: Name under which to register the block.
        """

        def decorator(block_class):
            cls._registry[block_name] = block_class
            logger.debug(
                f"Registered block '{block_name}' with class '{block_class.__name__}'"
            )
            return block_class

        return decorator

    @classmethod
    def get_registry(cls):
        """
        Retrieve the current registry map of block types.

        :return: Dictionary of registered block names and classes.
        """
        logger.debug("Fetching the block registry map.")
        return cls._registry


class PromptRegistry:
    """Registry for managing Jinja2 prompt templates."""

    _registry: Dict[str, Template] = {}

    @classmethod
    def register(cls, name: str):
        """Decorator to register a Jinja2 template function by name.

        :param name: Name of the template to register.
        :return: A decorator that registers the Jinja2 template function.
        """

        def decorator(func):
            template_str = func()
            cls._registry[name] = Template(template_str)
            logger.debug(f"Registered prompt template '{name}'")
            return func

        return decorator

    @classmethod
    def get_template(cls, name: str) -> Template:
        """Retrieve a Jinja2 template by name.

        :param name: Name of the template to retrieve.
        :return: The Jinja2 template instance.
        """
        if name not in cls._registry:
            raise KeyError(f"Template '{name}' not found.")
        logger.debug(f"Retrieving prompt template '{name}'")
        return cls._registry[name]

    @classmethod
    def get_registry(cls):
        """
        Retrieve the current registry map of block types.

        :return: Dictionary of registered block names and classes.
        """
        logger.debug("Fetching the block registry map.")
        return cls._registry

    @classmethod
    def render_template(
        cls,
        name: str,
        messages: Union[str, List[Dict[str, str]]],
        add_generation_prompt: bool = True,
    ) -> str:
        """Render the template with the provided messages or query.

        :param name: Name of the template to render.
        :param messages: Either a single query string or a list of messages (each as a dict with 'role' and 'content').
        :param add_generation_prompt: Whether to add a generation prompt at the end.
        :return: The rendered prompt as a string.
        """

        # Special handling for "blank" template
        if name == "blank":
            if not isinstance(messages, str):
                raise ValueError(
                    "The 'blank' template can only be used with a single query string, not a list of messages."
                )
            return messages  # Return the query as-is without templating

        # Get the template
        template = cls.get_template(name)

        # If `messages` is a string, wrap it in a list with a default user role
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        # Render the template with the `messages` list
        return template.render(
            messages=messages, add_generation_prompt=add_generation_prompt
        )



# Third Party
from datasets import load_dataset
from openai import OpenAI
import click

# First Party
from sdg_hub.flow import  Flow
from sdg_hub.logger_config import setup_logger
from sdg_hub.pipeline import Pipeline
from sdg_hub.sdg import SDG
from sdg_hub.prompts import PromptRegistry
from sdg_hub.blocks import BlockRegistry, Block
from transformers import AutoTokenizer
import re
from typing import List
from datasets import Dataset

logger = setup_logger(__name__)

### Nemotron Chat Template with detailed thinking on
@PromptRegistry.register("nvidia/Llama-3_3-Nemotron-Super-49B-v1")
def nemotron_chat_template():
    return """{{- bos_token }}
{{- "<|start_header_id|>system<|end_header_id|>\n\n" }}detailed thinking on{{- "<|eot_id|>" }}
{%- for message in messages %}
  {%- if message['role'] == 'assistant' and '</think>' in message['content'] %}
    {%- set content = message['content'].split('</think>')[-1].lstrip() %}
  {%- else %}
    {%- set content = message['content'] %}
  {%- endif %}
  {{- '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n' + content | trim + '<|eot_id|>' }}
{%- endfor %}
{%- if add_generation_prompt %}
  {{- '<|start_header_id|>assistant<|end_header_id|>\n\n' }}
{%- endif %}"""


@BlockRegistry.register("PostProcessThinkingBlock")
class PostProcessThinkingBlock(Block):
    def __init__(self, block_name: str, column_name: str) -> None:
        super().__init__(block_name=block_name)  
        self.column_name = column_name
    
    
    def generate(self, samples: Dataset):
        def post_process_thinking(x):
            if '</think>' in x[self.column_name]:
                x[self.column_name] = x[self.column_name].split('</think>')[-1].lstrip()
            return x
        samples = samples.map(post_process_thinking)
        return samples

@BlockRegistry.register("RegexParserBlock")
class RegexParserBlock(Block):
    def __init__(self, block_name: str, column_name: str, parsing_pattern: str="", parser_cleanup_tags: List[str]=[], output_cols: List[str]=[]) -> None:
        super().__init__(block_name=block_name)
        self.column_name = column_name
        self.parsing_pattern = parsing_pattern
        self.parser_cleanup_tags = parser_cleanup_tags
        self.output_cols = output_cols

    def generate(self, samples: Dataset):
        
        if self.parsing_pattern:
            new_data = []
            for sample in samples:
                parsed_outputs = self._parse(sample[self.column_name])
                
                max_length = max(len(value) for value in parsed_outputs.values())
                for values in zip(*(lst[:max_length] for lst in parsed_outputs.values())):
                    new_data.append({**sample, **dict(zip(parsed_outputs.keys(), values))})
            samples = Dataset.from_list(new_data)
        if self.parser_cleanup_tags:
            for clean_tag in self.parser_cleanup_tags:
               samples = samples.map(lambda x: {column_name: x[column_name].replace(clean_tag, "") for column_name in self.output_cols})
        return samples

    def _parse(self, generated_string):      
        pattern = re.compile(self.parsing_pattern, re.DOTALL)
        all_matches = pattern.findall(generated_string)
        matches = {column_name: [] for column_name in self.output_cols}
        if all_matches and isinstance(all_matches[0], tuple):
            for match in all_matches:
                for column_name, value in zip(self.output_cols, match):
                    value = value.strip()
                    # for clean_tag in self.parser_cleanup_tags:
                    #     value = value.replace(clean_tag, "")
                    matches[column_name].append(value)
        else:
            matches[self.output_cols[0]] = (
                [match.strip() for match in all_matches] if all_matches else []
            )
        return matches

@click.command()
@click.option(
    "--ds_path",
    type=click.Path(exists=True),
    required=True,
    help="Path to the dataset.",
)
@click.option("--bs", type=int, default=8, show_default=True, help="Batch size.")
@click.option(
    "--num_workers", type=int, default=32, show_default=True, help="Number of workers."
)
@click.option(
    "--save_path", type=click.Path(), required=True, help="Path to save the output."
)
@click.option(
    "--endpoint", type=str, required=True, help="Endpoint for data processing."
)
@click.option(
    "--flow", type=str, required=True, help="Flow configuration for the process."
)
@click.option(
    "--checkpoint_dir",
    type=click.Path(),
    required=True,
    help="Path to save checkpoints.",
)
@click.option(
    "--save_freq",
    type=int,
    default=2,
    show_default=True,
    help="Frequency to save checkpoints.",
)
@click.option("--debug", is_flag=True, help="Enable debug mode.")
@click.option("--dataset_start_index", type=int, default=0, help="Start index of the dataset.")
@click.option("--dataset_end_index", type=int, default=None, help="End index of the dataset.")
def main(
    ds_path,
    bs,
    num_workers,
    save_path,
    endpoint,
    flow,
    checkpoint_dir,
    save_freq,
    debug,
    dataset_start_index,
    dataset_end_index,
):
    """
    Main function to process the dataset.

    Parameters:
    ds_path (str): Path to the dataset.
    bs (int): Batch size.
    num_workers (int): Number of workers.
    save_path (str): Path to save the output.
    endpoint (str): Endpoint for data processing.
    flow (str): Flow configuration for the process.
    checkpoint_dir (str): Path to save checkpoints.
    save_freq (int): Frequency to save checkpoints.
    debug (bool): Enable debug mode.
    """
    logger.info(f"Generation configuration: {locals()}\n\n")
    ds = load_dataset("json", data_files=ds_path, split="train")
    if dataset_start_index is not None and dataset_end_index is not None:
        if dataset_end_index > len(ds):
            dataset_end_index = len(ds)
        ds = ds.select(range(dataset_start_index, dataset_end_index))
        logger.info(f"Dataset sliced from {dataset_start_index} to {dataset_end_index}")

    if debug:
        # For debugging, use a smaller subset of the dataset
        ds = ds.shuffle(seed=42).select(range(30))

    openai_api_key = "EMPTY"
    openai_api_base = endpoint

    client = OpenAI(
        api_key=openai_api_key,
        base_url=openai_api_base,
    )

    flow_cfg = Flow(client).get_flow_from_file(flow)
    sdg = SDG(
        [Pipeline(flow_cfg)],
        num_workers=num_workers,
        batch_size=bs,
        save_freq=save_freq,
    )
    generated_data = sdg.generate(ds, checkpoint_dir=checkpoint_dir)
    
    save_path = save_path.replace(".jsonl", f"_{dataset_start_index}_{dataset_end_index}.jsonl")
    generated_data.to_json(save_path, orient="records", lines=True)
    logger.info(f"Data saved to {save_path}")


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



# Local
from .registry import PromptRegistry


@PromptRegistry.register("blank")
def blank_chat_template():
    return """{{ messages }}"""


@PromptRegistry.register("instructlab")
def instructlab_chat_template():
    return """{% for message in messages %}{% if message['role'] == 'pretraining' %}{{ '<|pretrain|>' + message['content'] + '<|endoftext|>' + '<|/pretrain|>' }}{% elif message['role'] == 'system' %}{{ '<|system|>' + '\n' + message['content'] + '\n' }}{% elif message['role'] == 'user' %}{{ '<|user|>' + '\n' + message['content'] + '\n' }}{% elif message['role'] == 'assistant' %}{{ '<|assistant|>' + '\n' + message['content'] + '<|endoftext|>' + ('' if loop.last else '\n') }}{% endif %}{% if loop.last and add_generation_prompt %}{{ '<|assistant|>' + '\n' }}{% endif %}{% endfor %}"""


@PromptRegistry.register("mistralai")
def mistral_chat_template():
    return """{%- if messages[0]['role'] == 'system' %}\n    {%- set system_message = messages[0]['content'] %}\n    {%- set loop_messages = messages[1:] %}\n{%- else %}\n    {%- set loop_messages = messages %}\n{%- endif %}\n\n<s>\n{%- for message in loop_messages %}\n    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}\n        {{- raise_exception('After the optional system message, conversation roles must alternate user/assistant/user/assistant/...') }}\n    {%- endif %}\n    {%- if message['role'] == 'user' %}\n        {%- if loop.first and system_message is defined %}\n            {{- ' [INST] ' + system_message + '\\n\\n' + message['content'] + ' [/INST]' }}\n        {%- else %}\n            {{- ' [INST] ' + message['content'] + ' [/INST]' }}\n        {%- endif %}\n    {%- elif message['role'] == 'assistant' %}\n        {{- ' ' + message['content'] + '</s>'}}\n    {%- else %}\n        {{- raise_exception('Only user and assistant roles are supported, with the exception of an initial optional system message!') }}\n    {%- endif %}\n{%- endfor %}\n"""



# Third Party
from datasets import concatenate_datasets


def safe_concatenate_datasets(datasets: list):
    """
    Concatenate datasets safely, ignoring any datasets that are None or empty.
    """
    filtered_datasets = [ds for ds in datasets if ds is not None and ds.num_rows > 0]

    if not filtered_datasets:
        return None

    return concatenate_datasets(filtered_datasets)



# SPDX-License-Identifier: Apache-2.0
# Standard
from collections import Counter
from typing import Any, Dict, List
import json
import re

# Third Party
from datasets import Dataset
from jinja2 import Template
import openai

# Local
from .block import Block
from ..logger_config import setup_logger
from ..registry import BlockRegistry, PromptRegistry

logger = setup_logger(__name__)


def server_supports_batched(client, model_id: str) -> bool:
    supported = getattr(client, "server_supports_batched", None)
    if supported is not None:
        return supported
    try:
        # Make a test call to the server to determine whether it supports
        # multiple input prompts per request and also the n parameter
        response = client.completions.create(
            model=model_id, prompt=["test1", "test2"], max_tokens=1, n=3
        )
        # Number outputs should be 2 * 3 = 6
        supported = len(response.choices) == 6
    except openai.InternalServerError:
        supported = False
    client.server_supports_batched = supported
    logger.info(f"LLM server supports batched inputs: {client.server_supports_batched}")
    return supported


@BlockRegistry.register("LLMBlock")
# pylint: disable=dangerous-default-value
class LLMBlock(Block):
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        block_name,
        config_path,
        client,
        output_cols,
        parser_kwargs={},
        model_prompt="{prompt}",
        model_id=None,
        **batch_kwargs,
    ) -> None:
        super().__init__(block_name)
        self.block_config = self._load_config(config_path)
        self.prompt_struct = (
            """{system}\n{introduction}\n{principles}\n{examples}\n{generation}"""
        )
        filtered_config = {
            k: (v if v is not None else "") for k, v in self.block_config.items()
        }
        self.prompt_template = Template(self.prompt_struct.format(**filtered_config))
        self.client = client
        if model_id:
            self.model = model_id
        else:
            # get the default model id from client
            self.model = self.client.models.list().data[0].id

        self.model_prompt = model_prompt
        self.output_cols = output_cols
        self.batch_params = batch_kwargs.get("batch_kwargs", {})
        self.parser_name = parser_kwargs.get("parser_name", None)
        self.parsing_pattern = parser_kwargs.get("parsing_pattern", None)
        self.parser_cleanup_tags = parser_kwargs.get("parser_cleanup_tags", None)
        self.defaults = {
            "model": self.model,
            "temperature": 0,
            "max_tokens": 4096,
        }

        # Whether the LLM server supports a list of input prompts
        # and supports the n parameter to generate n outputs per input
        self.server_supports_batched = server_supports_batched(client, self.model)

    def _parse(self, generated_string) -> dict:
        matches = {}

        if self.parser_name is not None and self.parser_name == "custom":
            pattern = re.compile(self.parsing_pattern, re.DOTALL)
            all_matches = pattern.findall(generated_string)
            matches = {column_name: [] for column_name in self.output_cols}
            if all_matches and isinstance(all_matches[0], tuple):
                for match in all_matches:
                    for column_name, value in zip(self.output_cols, match):
                        value = value.strip()
                        for clean_tag in self.parser_cleanup_tags:
                            value = value.replace(clean_tag, "")
                        matches[column_name].append(value)
            else:
                matches[self.output_cols[0]] = (
                    [match.strip() for match in all_matches] if all_matches else []
                )
        else:
            for start_tag, end_tag, output_col in zip(
                self.block_config.get("start_tags", []),
                self.block_config.get("end_tags", []),
                self.output_cols,
            ):
                if not start_tag and not end_tag:
                    matches[output_col] = [
                        generated_string.strip() if generated_string else None
                    ]
                else:
                    pattern = re.escape(start_tag) + r"(.*?)" + re.escape(end_tag)
                    all_matches = re.findall(pattern, generated_string, re.DOTALL)
                    matches[output_col] = (
                        [match.strip() for match in all_matches] if all_matches else []
                    )

        return matches

    def _format_prompt(self, sample: Dict) -> str:
        prompt_templated_str = self.prompt_template.render(sample).strip()
        return PromptRegistry.render_template(
            self.model_prompt, prompt_templated_str, add_generation_prompt=True
        ).strip()

    def _generate(self, samples, **gen_kwargs) -> list:
        prompts = [self._format_prompt(sample) for sample in samples]
        logger.debug("Prompt: %s", prompts[0])
        generate_args = {**self.defaults, **gen_kwargs}

        if self.server_supports_batched:
            response = self.client.completions.create(prompt=prompts, **generate_args)
            # if stop is provided, then we need to add the stop token to the generated text,
            # this is because the stop token is not included in the generated text - this is a limitation of the openai api
            # we need to add the stop token to the generated text to make it consistent for the parser
            if "stop" in generate_args:
                return [
                    choice.text.strip() + "".join(generate_args["stop"])
                    for choice in response.choices
                ]
            return [choice.text.strip() for choice in response.choices]

        n = gen_kwargs.get("n", 1)
        results = []
        for prompt in prompts:
            for _ in range(n):
                response = self.client.completions.create(
                    prompt=prompt, **generate_args
                )
                if "stop" in generate_args:
                    results.append(
                        response.choices[0].text.strip()
                        + "".join(generate_args["stop"])
                    )
                results.append(response.choices[0].text.strip())
        return results

    def generate(self, samples: Dataset, **gen_kwargs) -> Dataset:
        """
        Generate the output from the block. This method should first validate the input data,
        then generate the output, and finally parse the generated output before returning it.

        :return: The parsed output after generation.
        """
        num_samples = self.block_config.get("num_samples", None)
        logger.debug("Generating outputs for {} samples".format(len(samples)))

        if (num_samples is not None) and ("num_samples" not in samples.column_names):
            samples = samples.add_column("num_samples", [num_samples] * len(samples))

        # validate each sample
        # Log errors and remove invalid samples
        valid_samples = []

        for sample in samples:
            if self._validate(self.prompt_template, sample):
                valid_samples.append(sample)
            else:
                logger.warning(
                    f"Sample failed validation: {sample}"
                )  # Log details of the failed sample

        samples = valid_samples

        if len(samples) == 0:
            logger.warning(
                "No valid samples to generate outputs for, returning empty dataset"
            )
            return Dataset.from_list([])

        # generate the output

        outputs = self._generate(samples, **gen_kwargs)

        logger.debug("Generated outputs: %s", outputs)

        num_parallel_samples = gen_kwargs.get("n", 1)
        extended_samples = []

        # Duplicate each input sample n times, where n is the number
        # of output sequences generated per input, so that we can
        # pair up the inputs and outputs.
        for item in samples:
            extended_samples.extend([item] * num_parallel_samples)

        new_data = []
        for sample, output in zip(extended_samples, outputs):
            parsed_outputs = self._parse(output)
            max_length = max(len(value) for value in parsed_outputs.values())
            for values in zip(*(lst[:max_length] for lst in parsed_outputs.values())):
                new_data.append({**sample, **dict(zip(parsed_outputs.keys(), values))})

        return Dataset.from_list(new_data)


@BlockRegistry.register("ConditionalLLMBlock")
class ConditionalLLMBlock(LLMBlock):
    def __init__(
        self,
        block_name,
        config_paths,
        client,
        model_id,
        output_cols,
        selector_column_name,
        model_prompt="{prompt}",
        **batch_kwargs,
    ) -> None:
        super().__init__(
            block_name=block_name,
            config_path=list(config_paths.values())[0],
            client=client,
            model_id=model_id,
            output_cols=output_cols,
            model_prompt=model_prompt,
            **batch_kwargs,
        )
        self.selector_column_name = selector_column_name
        self.prompt_template = {}
        if "All" in config_paths:
            self.prompt_template = self.prompt_struct.format(**self.block_config)
        else:
            for config_key, config in config_paths.items():
                # Template(self.prompt_struct.format(**filtered_config))
                filtered_config = {
                    k: (v if v is not None else "") for k, v in self.block_config.items()
                }
                self.prompt_template[config_key] = Template(self.prompt_struct.format(
                    **self._load_config(config)
                ))

    def _format_prompt(self, sample: Dict) -> str:
        if isinstance(self.prompt_template, dict):
            return (
                self.prompt_template[sample[self.selector_column_name]]
                .render(**sample)
                .strip()
            )

        return self.prompt_template.render(**sample).strip()

    def _validate(self, prompt_template: str, input_dict: Dict[str, Any]) -> bool:
        if isinstance(prompt_template, dict):
            prompt_template = prompt_template[input_dict[self.selector_column_name]]
        return super()._validate(prompt_template, input_dict)


@BlockRegistry.register("LLMLogProbBlock")
class LLMLogProbBlock(LLMBlock):
    # init with init of the parent class
    def __init__(
        self,
        block_name,
        config_path,
        client,
        output_cols,
        parser_kwargs={},
        model_prompt="{prompt}",
        model_id=None,
        **batch_kwargs,
    ) -> None:
        super().__init__(
            block_name=block_name,
            config_path=config_path,
            client=client,
            output_cols=output_cols,
            parser_kwargs=parser_kwargs,
            model_prompt=model_prompt,
            model_id=model_id,
            **batch_kwargs,
        )

    def _generate_logprobs(self, samples, **gen_kwargs):
        prompts = [
            self.model_prompt.format(prompt=self._format_prompt(sample))
            for sample in samples
        ]
        generate_args = {**self.defaults, **gen_kwargs}

        # verify if logprobs is mentioned in the generate_args, if not add it and return top10 logprobs
        if "logprobs" not in generate_args:
            generate_args["logprobs"] = 10

        if self.server_supports_batched:
            response = self.client.completions.create(prompt=prompts, **generate_args)
            return [choice.logprobs.top_logprobs for choice in response.choices]

        n = gen_kwargs.get("n", 1)
        results = []
        for prompt in prompts:
            for _ in range(n):
                response = self.client.completions.create(
                    prompt=prompt, **generate_args
                )
                results.append(response.choices[0].logprobs.top_logprobs)
        return results

    def _parse(self, generations: List[List[Dict]]) -> List[List[str]]:
        # override the parse method to convert the generations to json string
        # convert the generations to json string to save as dataset
        # this is because the dataset can only store key value pairs which are consistent
        return [[json.dumps(item) for item in sublist] for sublist in generations]

    def generate(self, samples: Dataset, **gen_kwargs) -> Dataset:
        """
        Generate the output from the block. This method should first validate the input data,
        then generate the output, and finally parse the generated output before returning it.

        :return: The parsed output after generation.
        """
        num_samples = self.block_config.get("num_samples", None)
        logger.debug("Generating outputs for {} samples".format(len(samples)))

        if (num_samples is not None) and ("num_samples" not in samples.column_names):
            samples = samples.add_column("num_samples", [num_samples] * len(samples))

        # validate each sample
        # Log errors and remove invalid samples
        valid_samples = []

        for sample in samples:
            if self._validate(self.prompt_template, sample):
                valid_samples.append(sample)
            else:
                logger.warning(
                    f"Sample failed validation: {sample}"
                )  # Log details of the failed sample

        samples = valid_samples

        if len(samples) == 0:
            logger.warning(
                "No valid samples to generate outputs for, returning empty dataset"
            )
            return Dataset.from_list([])

        # generate the output

        outputs = self._generate_logprobs(samples, **gen_kwargs)
        logger.debug("Generated outputs: %s", outputs)

        output_dataset = Dataset.from_list(samples)
        output_dataset = output_dataset.add_column(
            self.output_cols[0],
            self._parse(outputs),  # pylint: disable=no-value-for-parameter
        )

        return output_dataset


@BlockRegistry.register("LLMMessagesBlock")
class LLMMessagesBlock(Block):
    def __init__(
        self,
        block_name,
        client,
        input_col,
        output_col,
        model_prompt=None, 
        model_id=None,
        **batch_kwargs,
    ) -> None:
        self.block_name = block_name
        self.model_prompt = model_prompt
        self.batch_params = batch_kwargs.get("batch_kwargs", {})
        self.input_col = input_col
        self.output_col = output_col
        self.client = client

        if model_id:
            self.model = model_id
        else:
            self.model = self.client.models.list().data[0].id
        
        self.defaults = {
            "model": self.model,
            "temperature": 0,
            "max_tokens": 4096,
        }
        self.server_supports_batched = server_supports_batched(client, self.model)

    def _generate(self, samples, **gen_kwargs) -> list:
        generate_args = {**self.defaults, **gen_kwargs}

        if "n" in generate_args and generate_args.get("temperature", 0) <= 0:
            generate_args["temperature"] = 0.7
            logger.warning(
                "Temperature should be greater than 0 for n > 1, setting temperature to 0.7"
            )

        messages = samples[self.input_col]

        results = []
        n = gen_kwargs.get("n", 1)
        for message in messages:
            responses = self.client.chat.completions.create(messages=message, **generate_args)
            if n > 1:
                results.append([choice.message.content for choice in responses.choices])
            else:
                results.append(responses.choices[0].message.content)
        return results

    def generate(self, samples: Dataset, **gen_kwargs) -> Dataset:
        outputs = self._generate(samples, **gen_kwargs)
        samples = samples.add_column(self.output_col, outputs)
        return samples



"""Module containing blocks for scoring responses using Reward Models."""

# Standard
from typing import Dict, List
import json
from urllib.parse import urljoin

# Third Party
from datasets import Dataset
import requests

# Local
from .block import Block
from ..logger_config import setup_logger
from ..registry import BlockRegistry

logger = setup_logger(__name__)


@BlockRegistry.register("PRMBlock")
class PRMBlock(Block):
    """A block for scoring responses using a ProcessReward Model (PRM) via HTTP API.

    This block sends prompts and responses to a PRM endpoint and returns reward scores
    for each step in the response.
    """

    def __init__(
        self,
        block_name: str,
        host: str,
        port: int,
        model_name: str,
        prompt_col: str,
        response_col: str,
        output_col: str = "step_rewards",
        system_prompt: str = None,
        endpoint: str = "pooling",
        step_separator: str = "\n\n",
        step_fill_token: str = "<extra_0>",
    ) -> None:
        r"""Initialize the PRM (Process Reward Model) Block.

        Parameters
        ----------
        block_name : str
            Name of the block
        host : str
            Hostname of the PRM service (e.g., "0.0.0.0" or "localhost")
        port : int
            Port number the service is running on
        model_name : str
            Name of the PRM model to use
        prompt_col : str
            Column name containing the prompt
        response_col : str
            Column name containing the response
        output_col : str, optional
            Column name to store the reward scores, by default "step_rewards"
        system_prompt : str, optional
            Optional system prompt to use for scoring, by default None
        endpoint : str, optional
            API endpoint name, by default "pooling"
        step_separator : str, optional
            Separator between steps in the response, by default "\n\n"
        step_fill_token : str, optional
            Model specific fill token for steps in the response, by default "<extra_0>" used by Qwen2.5-Math-PRM
        """
        super().__init__(block_name)
        # Construct base URL from host and port
        self.base_url = f"http://{host.strip('/')}:{port}/"
        self.endpoint = endpoint.strip("/")

        # Construct the full API URL using urljoin
        self.api_url = urljoin(self.base_url, self.endpoint)
        logger.info(f"Initialized PRMBlock with API URL: {self.api_url}")

        self.model_name = model_name
        self.prompt_col = prompt_col
        self.response_col = response_col
        self.output_col = output_col
        self.system_prompt = system_prompt
        self.step_separator = step_separator
        self.step_fill_token = step_fill_token

    def _post_request(self, messages: List[Dict]) -> requests.Response:
        """Make POST request to PRM API endpoint.

        Parameters
        ----------
        messages : List[Dict]
            List of message dictionaries to send to the API

        Returns
        -------
        requests.Response
            Response from the API
        """
        headers = {"User-Agent": "PRMBlock Client"}
        prompt = {"model": self.model_name, "messages": messages}
        response = requests.post(self.api_url, headers=headers, json=prompt)
        return response

    def _format_messages(self, sample: Dict) -> List[Dict]:
        """Format input sample into messages for the PRM API.

        Parameters
        ----------
        sample : Dict
            Input sample containing prompt and response

        Returns
        -------
        List[Dict]
            Formatted messages for the API
        """
        messages = []
        if self.system_prompt:
            messages.append({"role": "system", "content": self.system_prompt})

        messages.append({"role": "user", "content": sample[self.prompt_col]})
        messages.append(
            {
                "role": "assistant",
                "content": self.step_fill_token.join(sample[self.response_col].split(self.step_separator))
                + self.step_fill_token,
            }
        )
        return messages

    def _extract_rewards(self, response: requests.Response) -> List[float]:
        """Extract reward scores from API response.

        Parameters
        ----------
        response : requests.Response
            Response from the API

        Returns
        -------
        List[float]
            List of reward scores
        """
        try:
            response_data = response.json()
            rewards = [x[1] for x in response_data["data"][0]["data"]]
            return rewards
        except (KeyError, IndexError, json.JSONDecodeError) as e:
            logger.error(f"Error extracting rewards from response: {e}")
            return []

    def _generate(self, sample: dict) -> dict:
        """Generate reward scores for the input samples.

        Parameters
        ----------
        sample : dict
            Input sample to score

        Returns
        -------
        dict
            Dictionary with added reward scores column
        """
        messages = self._format_messages(sample)
        rm_response = self._post_request(messages)

        if rm_response.status_code != 200:
            logger.error(f"API request failed with status {rm_response.status_code}")
            rewards = [0.0] * len(
                sample[self.response_col].split(self.step_separator)
            )  # Default to 0 scores on failure
        else:
            rewards = self._extract_rewards(rm_response)

        sample[self.output_col] = rewards
        return sample

    def generate(self, samples: Dataset, batch_size: int = 4) -> Dataset:
        """Generate reward scores for the input samples.

        Parameters
        ----------
        samples : Dataset
            Input dataset containing samples to score
        batch_size : int, optional
            Number of processes to use for parallel processing, by default 4

        Returns
        -------
        Dataset
            Dataset with added reward scores
        """
        return samples.map(self._generate, num_proc=batch_size)



# Standard
from unittest.mock import patch
import operator
import unittest

# Third Party
from datasets import Dataset, Features, Value

# First Party
from sdg_hub.blocks.filterblock import FilterByValueBlock


class TestFilterByValueBlock(unittest.TestCase):
    def setUp(self):
        self.block = FilterByValueBlock(
            filter_column="age",
            filter_value=30,
            operation=operator.eq,
            convert_dtype=int,
        )
        self.block_with_list = FilterByValueBlock(
            filter_column="age",
            filter_value=[30, 35],
            operation=operator.eq,
            convert_dtype=int,
        )
        self.dataset = Dataset.from_dict(
            {"age": ["25", "30", "35", "forty", "45"]},
            features=Features({"age": Value("string")}),
        )

    @patch("sdg_hub.blocks.filterblock.logger")
    def test_generate_mixed_types(self, mock_logger):
        filtered_dataset = self.block.generate(self.dataset)
        self.assertEqual(len(filtered_dataset), 1)
        self.assertEqual(filtered_dataset["age"], [30])
        mock_logger.error.assert_called()

    @patch("sdg_hub.blocks.filterblock.logger")
    def test_generate_mixed_types_multi_value(self, mock_logger):
        filtered_dataset = self.block_with_list.generate(self.dataset)
        self.assertEqual(len(filtered_dataset), 2)
        self.assertEqual(filtered_dataset["age"], [30, 35])
        mock_logger.error.assert_called()



# SPDX-License-Identifier: Apache-2.0
# Standard
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List
import traceback
import uuid

# Third Party
from datasets import Dataset, load_dataset
from datasets.data_files import EmptyDatasetError
from tqdm import tqdm

# Local
from .logger_config import setup_logger
from .pipeline import Pipeline
from .utils.datautils import safe_concatenate_datasets


logger = setup_logger(__name__)


class SDG:
    def __init__(
        self, pipelines: List[Pipeline], num_workers=1, batch_size=None, save_freq=None
    ) -> None:
        self.pipelines = pipelines
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.save_freq = save_freq

    def _split_dataset(self, dataset: Dataset, batch_size: int) -> List[Dataset]:
        """Split the dataset into smaller batches."""
        total_size = len(dataset)
        num_batches = (total_size + batch_size - 1) // batch_size

        batches = [
            (i * batch_size, min((i + 1) * batch_size, total_size))
            for i in tqdm(range(num_batches))
        ]

        return batches

    def _get_missing_data(self, seed_data, generated_data):
        # Get the common columns between the two datasets
        common_columns = list(
            set(seed_data.column_names) & set(generated_data.column_names)
        )

        # Extract the relevant data based on common columns
        seed_data_common = seed_data.select_columns(common_columns)
        generated_data_common = generated_data.select_columns(common_columns)

        # Convert to Pandas DataFrames for easier comparison
        seed_df = seed_data_common.to_pandas()
        generated_df = generated_data_common.to_pandas()

        # Identify missing rows
        missing_df = seed_df[
            ~seed_df.apply(tuple, 1).isin(generated_df.apply(tuple, 1))
        ]

        # Convert back to Dataset
        missing_data = Dataset.from_pandas(missing_df, preserve_index=False)

        return missing_data

    def _save_intermediate_checkpoint(self, dataset, checkpoint_dir):
        checkpoint_id = uuid.uuid4().hex
        checkpoint_file = f"{checkpoint_dir}/data_checkpoint_{checkpoint_id}.jsonl"
        logger.info(f"Saving checkpoint to {checkpoint_file}")
        dataset.to_json(checkpoint_file, orient="records", lines=True)

    @staticmethod
    def _generate_data(pipelines, input_split, ds, i=None):
        logger.info(f"Processing split {i}")
        input_split = ds.select(range(input_split[0], input_split[1]))
        try:
            for pipeline in pipelines:
                input_split = pipeline.generate(input_split)
            return input_split
        except Exception as e:
            logger.error(f"Error processing split {i}: {e}")
            traceback.print_exc()
            return None

    def generate(self, dataset: Dataset, checkpoint_dir=None) -> Dataset:
        # check if checkpoint_dir exists
        pre_generated_data = []
        if checkpoint_dir is not None:
            try:
                # check if there are any existing checkpoints
                pre_generated_data = load_dataset(
                    "json", data_dir=checkpoint_dir, split="train"
                )
                logger.info(
                    f"Loading existing checkpoints from {checkpoint_dir}, with {pre_generated_data.num_rows} rows"
                )
                seed_data = self._get_missing_data(dataset, pre_generated_data)
                if seed_data.num_rows == 0:
                    logger.info(
                        f"All seed data has been generated, no missing rows found, returning data from {checkpoint_dir}"
                    )
                    return pre_generated_data
                logger.info(f"Found {seed_data.num_rows} missing rows in the dataset")

            except EmptyDatasetError:
                logger.info(
                    f"No existing checkpoints found in {checkpoint_dir}, generating from scratch"
                )
                seed_data = dataset

        else:
            seed_data = dataset

        if not self.batch_size:
            # If batch size is not provided, generate the dataset in a single pass
            generated_dataset = seed_data
            # generated_data is initialized with seed_data, and it gets updated with each pipeline
            for pipeline in self.pipelines:
                generated_dataset = pipeline.generate(seed_data)
            return generated_dataset
        
        logger.info("Splitting the dataset into smaller batches")
        input_splits = (
            self._split_dataset(seed_data, self.batch_size)
            if self.batch_size
            else [seed_data]
        )
        logger.info(
            f"Generating dataset with {len(input_splits)} splits, batch size {self.batch_size}, and {self.num_workers} workers"
        )

        generated_data = [pre_generated_data] if pre_generated_data else []
        last_saved_split_index = 0  # To track the last saved split

        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            futures = [
                executor.submit(
                    self._generate_data, self.pipelines, input_split, seed_data, i
                )
                for i, input_split in enumerate(input_splits)
            ]

            for i, future in enumerate(tqdm(as_completed(futures), total=len(futures))):
                generated_data_split = future.result()  # Ensure each future completes

                if generated_data_split:
                    generated_data.append(generated_data_split)
                    logger.info(f"Finished future processing split {i} \n\n")
                    if self.save_freq and (i + 1) % self.save_freq == 0:
                        # Save only the new splits since the last checkpoint
                        new_splits = generated_data[last_saved_split_index : i + 1]
                        checkpoint_dataset = safe_concatenate_datasets(new_splits)
                        # check if checkpoint_dataset is not None
                        if checkpoint_dataset:
                            self._save_intermediate_checkpoint(
                                checkpoint_dataset, checkpoint_dir
                            )

                            last_saved_split_index = i + 1

        generated_dataset = safe_concatenate_datasets(generated_data)

        return generated_dataset
